#include <rpc/rpc.h>
#include <stdlib.h>
#include <string.h>
#include "offloading_rpc.h"
#include "../offloading.h"

static void* get_func(CONST char** cursor)
{
	CONST char* func_name = *cursor;
	int func_name_len = strlen(func_name);
	*cursor += func_name_len + 1;
	return find_func_by_name(func_name);
}

array* service(array* input, struct svc_req* req)
{
	array* result = malloc(sizeof(array));

	if (result == NULL)
	{
		printf("failed to allocate array\n");
		
		return NULL;
	}

	CONST char* cursor = input->buf;

	REMOTE_FUNC remote_func = (REMOTE_FUNC)get_func(&cursor);
	UNMARSHAL unmarshal = (UNMARSHAL)get_func(&cursor);
	MARSHAL marshal = (MARSHAL)get_func(&cursor);

	void* obj;
	unmarshal(&obj, cursor, input->len - (cursor - input->buf));

	void* output;
	remote_func(obj, &output);
	free(obj);

	char* result_buf;
	int result_size;
	marshal(result, &(result->buf), &(result->len));

	return result;
}

typedef array* (*LOCAL)(array *, struct svc_req *);

static void dispatcher(struct svc_req *rqstp, register SVCXPRT *transp)
{
	array argument = {0, 0};

	array* result;
	xdrproc_t _xdr_argument, _xdr_result;
	LOCAL local_func;

	switch (rqstp->rq_proc) {
	case REMOTE_CALL:
		_xdr_argument = (xdrproc_t) xdr_buffer;
		_xdr_result = (xdrproc_t) xdr_buffer;
		local_func = (LOCAL)service;
		break;
	
	default:
		(void) svc_sendreply (transp, (xdrproc_t) xdr_void, (char *)NULL);
		return;
	}
	
	if (!svc_getargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
		svcerr_decode (transp);
		return;
	}
	
	result = (*local_func)((array *)&argument, rqstp);
	
	if (result != NULL && !svc_sendreply(transp, (xdrproc_t)_xdr_argument, (char*)result)) {
		svcerr_systemerr (transp);
	}

	free_array(result);
	
	if (!svc_freeargs (transp, (xdrproc_t) _xdr_argument, (caddr_t) &argument)) {
		fprintf (stderr, "%s", "unable to free arguments");
		exit (1);
	}
	
	return;
}


void server_init()
{
	register SVCXPRT *transp;
	
	pmap_unset (REMOTE_CALL_PROGRAM, REMOTE_CALL_VERSION);
	transp = svcudp_create(RPC_ANYSOCK);
	if (transp == NULL) {
		fprintf (stderr, "%s", "cannot create udp service.");
		return;
	}
	
	if (!svc_register(transp, REMOTE_CALL_PROGRAM, REMOTE_CALL_VERSION, dispatcher, IPPROTO_UDP)) {
		fprintf (stderr, "%s", "unable to register (REMOTE_CALL_PROGRAM, REMOTE_CALL_VERSION, udp).");
		return;
	}
	
 	transp = svctcp_create(RPC_ANYSOCK, 0, 0);
	if (transp == NULL) {
		fprintf (stderr, "%s", "cannot create tcp service.");
		return;
	}
	
	if (!svc_register(transp, REMOTE_CALL_PROGRAM, REMOTE_CALL_VERSION, dispatcher, IPPROTO_TCP)) {
		fprintf (stderr, "%s", "unable to register (REMOTE_CALL_PROGRAM, REMOTE_CALL_VERSION, tcp).");
		return;
	}

	svc_run ();
}

